import os

_CURRENT_DIR = os.path.dirname(__file__)


class Bitmaps:
    APPLE = os.path.join(_CURRENT_DIR, 'bitmaps/apple.bitmap.json')  # noqa
    APPLE_OLD = os.path.join(_CURRENT_DIR, 'bitmaps/apple_old.bitmap.json')  # noqa
    FACEBOOK = os.path.join(_CURRENT_DIR, 'bitmaps/facebook.bitmap.json')  # noqa
    GMAIL = os.path.join(_CURRENT_DIR, 'bitmaps/gmail.bitmap.json')  # noqa
    GOOGLE = os.path.join(_CURRENT_DIR, 'bitmaps/google.bitmap.json')  # noqa
    ITUNES = os.path.join(_CURRENT_DIR, 'bitmaps/itunes.bitmap.json')  # noqa
    PLUG = os.path.join(_CURRENT_DIR, 'bitmaps/plug.bitmap.json')  # noqa
    SMILEY = os.path.join(_CURRENT_DIR, 'bitmaps/smiley.bitmap.json')  # noqa
    WINDOWS = os.path.join(_CURRENT_DIR, 'bitmaps/windows.bitmap.json')  # noqa
    XBOX = os.path.join(_CURRENT_DIR, 'bitmaps/xbox.bitmap.json')  # noqa
    YOUTUBE = os.path.join(_CURRENT_DIR, 'bitmaps/youtube.bitmap.json')  # noqa


class Movies:
    PING_PONG = os.path.join(_CURRENT_DIR, 'movies/ping_pong.movie.json')  # noqa
